#!/usr/bin/perl -w
# raytracing module on exported ascii raster
# used by raytrace.sh
# (c) dimitrij mlekuz 2002
# released under GPL
# 
#   _rt.pl infile outfile x y
# where:
# infile is exported grass ascii raster layer
# outfile is result, ascii grass raster layer
# x,y are coordinates of sound source
#

sub readraster {
#reads raster and returns meta and data
# readraster "dmr", @meta, @data

  $InputFile=$_[0];

  #open input file
  open(INFILE, $InputFile) or die "File $InputFile does not exist.\n";

  #read metainformation
  $_=<INFILE>;
  @v=split(/ /, $_);
  $north=$v[1];

  $_=<INFILE>;
  @v=split(/ /, $_);
  $south=$v[1];

  $_=<INFILE>;
  @v=split(/ /, $_);
  $east=$v[1];

  $_=<INFILE>;
  @v=split(/ /, $_);
  $west=$v[1];

  $_=<INFILE>;
  @v=split(/ /, $_);
  $rows=$v[1];

  $_=<INFILE>;
  @v=split(/ /, $_);
  $cols=$v[1];

  $NSres=abs($north-$south)/$rows;
  $EWres=abs($east-$west)/$cols;

  @rasterMeta=($north,$south,$east,$west,$rows,$cols,$NSres,$EWres);

  #now, lets read data
  @rasterData=();

  #loop - read data values
  while(<INFILE>) {
    chomp($_);
    @values=split(/ /, $_);
    push @rasterData, [ @values ];     
   
  }#while

  close(INFILE);
} #readraster
#########################################################

sub profile {
#profiles raster map from x0,y0 to x1,y1
    my($x0,$y0,$x1,$y1)=@_;

    my($dx,$dy,$realX0,$realY0,$realX,$realY,$incx,$incy);
    my($disx,$disy,$hdist);

    $dx=$x1-$x0;
    $dy=$y1-$y0;

    if($dx==0 and $dy==0) { @rasterProfile=(); return; }

    #       west                   EWres
    $realX0=$rasterMeta[3]+($x0-1)*$rasterMeta[7];
    #       north                  NSres
    $realY0=$rasterMeta[0]-($y0-1)*$rasterMeta[6];


    #print $realX0, " ", $realY0,"\n";
    @rasterProfile=();

 
    if($dx==0) { $incy=($y1-$y0)/abs($y1-$y0);  }
    else { $incy=$dy/abs($dx);  }

    if($dy==0) { $incx=($x1-$x0)/abs($x1-$x0);  }
    else { $incx=$dx/abs($dy);  }

    if(abs($incx)>1){ $incx=$incx/abs($incx); }
    if(abs($incy)>1){ $incy=$incy/abs($incy); }

    $x=$x0;
    $y=$y0;

    $disx=abs($x-$x1);
    $disy=abs($y-$y1);

    
   
    while ( $disx>abs($incx) or  $disy>abs($incy) ) {
        #print $disy," ",abs(int($incy)), "\n";
        #print $incx," ", $incy,"\n";
        $realX=$rasterMeta[3]+($x-1)*$rasterMeta[7];
        $realY=$rasterMeta[0]-($y-1)*$rasterMeta[6];

        $hdist=sqrt(($realX-$realX0)*($realX-$realX0)+($realY-$realY0)*($realY-$realY0));
        # print int($x), "** ",int($y),"\n";

        #print $hdist," ** ",$rasterData[$x-1][$y-1],"\n";

        push @rasterProfile, [ $hdist, $rasterData[$x-1][$y-1] ];

          
        $disx=abs($x-$x1);
        $disy=abs($y-$y1);
   
        if($disx>abs(int($incx))) { $x=$x+$incx; }
        if($disy>abs(int($incy))) { $y=$y+$incy; }  

       
     
       
    } #while
   #lets enter last one
        $realX=$rasterMeta[3]+($x1-1)*$rasterMeta[7];
        $realY=$rasterMeta[0]-($y1-1)*$rasterMeta[6];

        $hdist=sqrt(($realX-$realX0)*($realX-$realX0)+($realY-$realY0)*($realY-$realY0));
        # print $x, " ", $y, " ",$realX, " ", $realY," ",$hdist,"\n";

        #print $hdist," ** ",$rasterData[$x1-1][$y1-1],"\n";
        push @rasterProfile, [ $hdist, $rasterData[$x1-1][$y1-1] ];

    #lets get lost of duplicates (there shouldn't be any!)
       #   undef %saw;
       #   @out = grep(!$saw{$_}++, @rasterProfile);
       #   @rasterProfile=@out;
    #print "done\n";
} #profile

#########################################################
sub findpeak {
    my($start)=$_[0];
    my($i)=0;
    my($fD)=$rasterProfile[$start][0];
    my($fZ)=$rasterProfile[$start][1];
    #print $firstZ,"\n";
    my($maxAngle)=-99999;
    my($next)=$start+1;
    for($i=$start+1;$i<@rasterProfile;$i++) {
      my($cD)=$rasterProfile[$i][0];
      my($cZ)=$rasterProfile[$i][1];
      #print $cZ,"\n";
      my($angle)=($cZ-$fZ)/($cD-$fD);
      if ($angle>$maxAngle) { 
	  $maxAngle=$angle;
          $next=$i;
      }#if
    }#for

    return $next;
}
#########################################################
sub Fdist {
    $firstZ=$rasterProfile[0][1];
    $lastD=$rasterProfile[@rasterProfile-1][0];
    $lastZ=$rasterProfile[@rasterProfile-1][1];
    $directD=sqrt($lastD*$lastD+($firstZ-$lastZ)*($firstZ-$lastZ));
    #print $directD,"= directd\n";

    $currentD=0;
    $in=0;
    $n=0;
    while ( $in<@rasterProfile-1 ) {
        $n=findpeak($in);
        #print $in,"--->",$n," = ";
        
        $id=$rasterProfile[$in][0];
        $iz=$rasterProfile[$in][1];
        $nd=$rasterProfile[$n][0];
        $nz=$rasterProfile[$n][1];
        
        $dst=sqrt(($id-$nd)*($id-$nd)+($iz-$nz)*($iz-$nz));
        #print $dst,"\n";
        $currentD=$currentD+$dst;
        $in=$n;
    }
    if ((@rasterProfile-1)==0) { $currentD=$directD; }  
    return  $currentD-$directD;    
        
} #Fdist

#################################################################
#                                                               #
#################################################################
    

$noa=@ARGV;
if($noa!=4) { die "Usage: rt.pl infile outfile x y \n"; }

$InputFile=$ARGV[0];
$OutputFile=$ARGV[1];
$XX=$ARGV[2];
$YY=$ARGV[3];
 



@rasterMeta=();
@rasterData=();
@rasterProfile=();


readraster $InputFile;
;


$xres=abs($rasterMeta[2]-$rasterMeta[3])/$rasterMeta[5];
$yres=abs($rasterMeta[0]-$rasterMeta[1])/$rasterMeta[4];

$Ccol=int(($XX - $rasterMeta[3])/$xres+0.5);
$Rrow=$rasterMeta[4]-int(($YY - $rasterMeta[1])/$yres+0.5);

#print $xres,"   ", $yres, "  ",$Rrow,"   ",$Ccol;



open(OUTFILE, ">$OutputFile") or die "File can not be created.\n";

print OUTFILE  "north: ", $rasterMeta[0];
print OUTFILE  "south: ", $rasterMeta[1];
print OUTFILE  "east: ", $rasterMeta[2];
print OUTFILE  "west: ", $rasterMeta[3];
print OUTFILE  "rows: ", $rasterMeta[4];
print OUTFILE  "cols: ", $rasterMeta[5];

for($r=1; $r<=$rasterMeta[4]; $r++) {
    for($c=1; $c<=$rasterMeta[5]; $c++) {
   #   print $r," *  ",$c,"\n";  
     profile $Rrow,$Ccol,$r,$c;
     printf OUTFILE ("%g ",Fdist);
# if($c==$Ccol|| $r==$Rrow)
# {    printf OUTFILE ("%g ",666); print "yaba yaba do";}
#else
#{  profile $Rrow,$Ccol,$r,$c;
#    printf OUTFILE ("%g ",Fdist);
#}

    
  }
 printf OUTFILE ("\n");
}

close(OUTFILE);


#profile 50,50,50,51;
#print Fdist;








